import argparse
import asyncio
import atexit
import signal
import json
import platform
import os
import time
import traceback
import uuid
import numpy as np
from tqdm import tqdm
from net.train.dataset import load_dataset, iterate_batches
from net.networking.manual.manual_discovery import ManualDiscovery
from net.orchestration.node import Node
from net.networking.grpc.grpc_server import GRPCServer
from net.networking.udp.udp_discovery import UDPDiscovery
from net.networking.tailscale.tailscale_discovery import TailscaleDiscovery
from net.networking.grpc.grpc_peer_handle import GRPCPeerHandle
from net.topology.ring_memory_weighted_partitioning_strategy import RingMemoryWeightedPartitioningStrategy
from net.api import ChatGPTAPI
from net.download.shard_download import ShardDownloader, NoopShardDownloader
from net.download.download_progress import RepoProgressEvent
from net.download.new_shard_download import new_shard_downloader, has_exo_home_read_access, has_exo_home_write_access, ensure_exo_home, seed_models
from net.helpers import print_yellow_exo, find_available_port, DEBUG, get_system_info, get_or_create_node_id, get_all_ip_addresses_and_interfaces, terminal_link, shutdown
from net.inference.shard import Shard
from net.inference.inference_engine import get_inference_engine
from net.inference.tokenizers import resolve_tokenizer
from net.models import build_base_shard, get_repo
from net.viz.topology_viz import TopologyViz
import uvloop
import concurrent.futures
import resource
import psutil

# TODO: figure out why this is happening
os.environ["GRPC_VERBOSITY"] = "error"
os.environ["TRANSFORMERS_VERBOSITY"] = "error"
os.environ["TOKENIZERS_PARALLELISM"] = "true"

# Configure uvloop for maximum performance
def configure_uvloop():
    uvloop.install()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    # Increase file descriptor limits on Unix systems
    if not psutil.WINDOWS:
      soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
      try: resource.setrlimit(resource.RLIMIT_NOFILE, (hard, hard))
      except ValueError:
        try: resource.setrlimit(resource.RLIMIT_NOFILE, (8192, hard))
        except ValueError: pass

    loop.set_default_executor(concurrent.futures.ThreadPoolExecutor(max_workers=min(32, (os.cpu_count() or 1) * 4)))
    return loop

# parse args
parser = argparse.ArgumentParser(prog="hanzo-net", description="Hanzo Network - Distributed AI Compute")
parser.add_argument("command", nargs="?", choices=["run", "eval", "train"], help="Command to run")
parser.add_argument("model_name", nargs="?", help="Model name to run")
parser.add_argument("--default-model", type=str, default=None, help="Default model")
parser.add_argument("--iters", type=int, default=100, help="Training iterations")
parser.add_argument("--save-every", type=int, default=5, help="Save the model every N iterations.")
parser.add_argument("--data", type=str, default="exo/train/data/lora", help="Directory where training data lives")
parser.add_argument("--batch-size", type=int, default=1, help="Minibatch size.")
parser.add_argument("--resume-checkpoint", type=str, default=None, help="Path to a custom checkpoint to load")
parser.add_argument("--save-checkpoint-dir", type=str, default="checkpoints", help="Path to a folder where checkpoints are stored")
parser.add_argument("--node-id", type=str, default=None, help="Node ID")
parser.add_argument("--node-host", "--host", type=str, default=os.environ.get("HOST", "0.0.0.0"), help="Node host (can also use HOST env var)")
parser.add_argument("--node-port", "--port", type=int, default=os.environ.get("PORT", None), help="Node port (can also use PORT env var)")
parser.add_argument("--models-seed-dir", type=str, default=None, help="Model seed directory")
parser.add_argument("--listen-port", type=int, default=5678, help="Listening port for discovery")
parser.add_argument("--download-quick-check", action="store_true", help="Quick check local path for model shards download")
parser.add_argument("--max-parallel-downloads", type=int, default=8, help="Max parallel downloads for model shards download")
parser.add_argument("--broadcast-port", type=int, default=5678, help="Broadcast port for discovery")
parser.add_argument("--discovery-module", type=str, choices=["udp", "tailscale", "manual"], default="udp", help="Discovery module to use")
parser.add_argument("--discovery-timeout", type=int, default=30, help="Discovery timeout in seconds")
parser.add_argument("--discovery-config-path", type=str, default=None, help="Path to discovery config json file")
parser.add_argument("--wait-for-peers", type=int, default=0, help="Number of peers to wait to connect to before starting")
parser.add_argument("--chatgpt-api-port", type=int, default=52415, help="ChatGPT API port")
parser.add_argument("--chatgpt-api-response-timeout", type=int, default=900, help="ChatGPT API response timeout in seconds")
parser.add_argument("--max-generate-tokens", type=int, default=10000, help="Max tokens to generate in each request")
parser.add_argument("--inference-engine", type=str, default=None, help="Inference engine to use (mlx, tinygrad, or dummy)")
parser.add_argument("--disable-tui", action=argparse.BooleanOptionalAction, help="Disable TUI")
parser.add_argument("--run-model", type=str, help="Specify a model to run directly")
parser.add_argument("--prompt", type=str, help="Prompt for the model when using --run-model", default="Who are you?")
parser.add_argument("--default-temp", type=float, help="Default token sampling temperature", default=0.0)
parser.add_argument("--tailscale-api-key", type=str, default=None, help="Tailscale API key")
parser.add_argument("--tailnet-name", type=str, default=None, help="Tailnet name")
parser.add_argument("--node-id-filter", type=str, default=None, help="Comma separated list of allowed node IDs (only for UDP and Tailscale discovery)")
parser.add_argument("--interface-type-filter", type=str, default=None, help="Comma separated list of allowed interface types (only for UDP discovery)")
parser.add_argument("--system-prompt", type=str, default=None, help="System prompt for the ChatGPT API")
args = parser.parse_args()
if args.inference_engine:
    print(f"Selected inference engine: {args.inference_engine}")

print_yellow_exo()

system_info = get_system_info()
print(f"Detected system: {system_info}")

shard_downloader: ShardDownloader = new_shard_downloader(args.max_parallel_downloads) if args.inference_engine != "dummy" else NoopShardDownloader()
inference_engine_name = args.inference_engine or ("mlx" if system_info == "Apple Silicon Mac" else "tinygrad")
print(f"Inference engine name after selection: {inference_engine_name}")

inference_engine = get_inference_engine(inference_engine_name, shard_downloader)
print(f"Using inference engine: {inference_engine.__class__.__name__} with shard downloader: {shard_downloader.__class__.__name__}")

if args.node_port is None:
  args.node_port = find_available_port(args.node_host)
  if DEBUG >= 1: print(f"Using available port: {args.node_port}")

args.node_id = args.node_id or get_or_create_node_id()
chatgpt_api_endpoints = [f"http://{ip}:{args.chatgpt_api_port}/v1/chat/completions" for ip, _ in get_all_ip_addresses_and_interfaces()]
web_chat_urls = [f"http://{ip}:{args.chatgpt_api_port}" for ip, _ in get_all_ip_addresses_and_interfaces()]
if DEBUG >= 0:
  print("Chat interface started:")
  for web_chat_url in web_chat_urls:
    print(f" - {terminal_link(web_chat_url)}")
  print("ChatGPT API endpoint served at:")
  for chatgpt_api_endpoint in chatgpt_api_endpoints:
    print(f" - {terminal_link(chatgpt_api_endpoint)}")

# Convert node-id-filter and interface-type-filter to lists if provided
allowed_node_ids = args.node_id_filter.split(',') if args.node_id_filter else None
allowed_interface_types = args.interface_type_filter.split(',') if args.interface_type_filter else None

if args.discovery_module == "udp":
  discovery = UDPDiscovery(
    args.node_id,
    args.node_port,
    args.listen_port,
    args.broadcast_port,
    lambda peer_id, address, description, device_capabilities: GRPCPeerHandle(peer_id, address, description, device_capabilities),
    discovery_timeout=args.discovery_timeout,
    allowed_node_ids=allowed_node_ids,
    allowed_interface_types=allowed_interface_types
  )
elif args.discovery_module == "tailscale":
  discovery = TailscaleDiscovery(
    args.node_id,
    args.node_port,
    lambda peer_id, address, description, device_capabilities: GRPCPeerHandle(peer_id, address, description, device_capabilities),
    discovery_timeout=args.discovery_timeout,
    tailscale_api_key=args.tailscale_api_key,
    tailnet=args.tailnet_name,
    allowed_node_ids=allowed_node_ids
  )
elif args.discovery_module == "manual":
  if not args.discovery_config_path:
    raise ValueError(f"--discovery-config-path is required when using manual discovery. Please provide a path to a config json file.")
  discovery = ManualDiscovery(args.discovery_config_path, args.node_id, create_peer_handle=lambda peer_id, address, description, device_capabilities: GRPCPeerHandle(peer_id, address, description, device_capabilities))
topology_viz = TopologyViz(chatgpt_api_endpoints=chatgpt_api_endpoints, web_chat_urls=web_chat_urls) if not args.disable_tui else None
node = Node(
  args.node_id,
  None,
  inference_engine,
  discovery,
  shard_downloader,
  partitioning_strategy=RingMemoryWeightedPartitioningStrategy(),
  max_generate_tokens=args.max_generate_tokens,
  topology_viz=topology_viz,
  default_sample_temperature=args.default_temp
)
server = GRPCServer(node, args.node_host, args.node_port)
node.server = server
api = ChatGPTAPI(
  node,
  node.inference_engine.__class__.__name__,
  response_timeout=args.chatgpt_api_response_timeout,
  on_chat_completion_request=lambda req_id, __, prompt: topology_viz.update_prompt(req_id, prompt) if topology_viz else None,
  default_model=args.default_model,
  system_prompt=args.system_prompt
)
buffered_token_output = {}
def update_topology_viz(req_id, tokens, __):
  if not topology_viz: return
  if not node.inference_engine.shard: return
  if node.inference_engine.shard.model_id == 'stable-diffusion-2-1-base': return
  if req_id in buffered_token_output: buffered_token_output[req_id].extend(tokens)
  else: buffered_token_output[req_id] = tokens
  topology_viz.update_prompt_output(req_id, node.inference_engine.tokenizer.decode(buffered_token_output[req_id]))
node.on_token.register("update_topology_viz").on_next(update_topology_viz)
def update_prompt_viz(request_id, opaque_status: str):
  if not topology_viz: return
  try:
    status = json.loads(opaque_status)
    if status.get("type") != "node_status" or status.get("status") != "start_process_prompt": return
    topology_viz.update_prompt(request_id, status.get("prompt", "corrupted prompt (this should never happen)"))
  except Exception as e:
    if DEBUG >= 2:
      print(f"Failed to update prompt viz: {e}")
      traceback.print_exc()
node.on_opaque_status.register("update_prompt_viz").on_next(update_prompt_viz)

def preemptively_load_shard(request_id: str, opaque_status: str):
  try:
    status = json.loads(opaque_status)
    if status.get("type") != "node_status" or status.get("status") != "start_process_prompt": return
    current_shard = node.get_current_shard(Shard.from_dict(status.get("shard")))
    if DEBUG >= 2: print(f"Preemptively starting download for {current_shard}")
    asyncio.create_task(node.inference_engine.ensure_shard(current_shard))
  except Exception as e:
    if DEBUG >= 2:
      print(f"Failed to preemptively start download: {e}")
      traceback.print_exc()
node.on_opaque_status.register("preemptively_load_shard").on_next(preemptively_load_shard)

last_events: dict[str, tuple[float, RepoProgressEvent]] = {}
def throttled_broadcast(shard: Shard, event: RepoProgressEvent):
  global last_events
  current_time = time.time()
  if event.status == "not_started": return
  last_event = last_events.get(shard.model_id)
  if last_event and last_event[1].status == "complete" and event.status == "complete": return
  if last_event and last_event[0] == event.status and current_time - last_event[0] < 0.2: return
  last_events[shard.model_id] = (current_time, event)
  asyncio.create_task(node.broadcast_opaque_status("", json.dumps({"type": "download_progress", "node_id": node.id, "progress": event.to_dict()})))
shard_downloader.on_progress.register("broadcast").on_next(throttled_broadcast)

async def run_model_cli(node: Node, model_name: str, prompt: str):
  inference_class = node.inference_engine.__class__.__name__
  shard = build_base_shard(model_name, inference_class)
  if not shard:
    print(f"Error: Unsupported model '{model_name}' for inference engine {inference_class}")
    return
  tokenizer = await resolve_tokenizer(get_repo(shard.model_id, inference_class))
  request_id = str(uuid.uuid4())
  callback_id = f"cli-wait-response-{request_id}"
  callback = node.on_token.register(callback_id)
  if topology_viz:
    topology_viz.update_prompt(request_id, prompt)
  prompt = tokenizer.apply_chat_template([{"role": "user", "content": prompt}], tokenize=False, add_generation_prompt=True)

  try:
    print(f"Processing prompt: {prompt}")
    await node.process_prompt(shard, prompt, request_id=request_id)

    tokens = []
    def on_token(_request_id, _tokens, _is_finished):
      tokens.extend(_tokens)
      return _request_id == request_id and _is_finished
    await callback.wait(on_token, timeout=300)

    print("\nGenerated response:")
    print(tokenizer.decode(tokens))
  except Exception as e:
    print(f"Error processing prompt: {str(e)}")
    traceback.print_exc()
  finally:
    node.on_token.deregister(callback_id)

def clean_path(path):
    """Clean and resolve path"""
    if path.startswith("Optional("):
        path = path.strip('Optional("').rstrip('")')
    return os.path.expanduser(path)

async def hold_outstanding(node: Node):
  while node.outstanding_requests:
    await asyncio.sleep(.5)
  return

async def run_iter(node: Node, shard: Shard, train: bool, data, batch_size=1):
  losses = []
  tokens = []
  for batch in tqdm(iterate_batches(data, batch_size), total=len(data) // batch_size):
    _, _, lengths = batch
    losses.append(np.sum(lengths * await node.enqueue_example(shard, *batch, train=train)))
    tokens.append(np.sum(lengths))
  total_tokens = np.sum(tokens)
  total_loss = np.sum(losses) / total_tokens

  return total_loss, total_tokens

async def eval_model_cli(node: Node, model_name, dataloader, batch_size, num_batches=-1):
  inference_class = node.inference_engine.__class__.__name__
  shard = build_base_shard(model_name, inference_class)
  if not shard:
    print(f"Error: Unsupported model '{model_name}' for inference engine {inference_class}")
    return
  tokenizer = await resolve_tokenizer(get_repo(shard.model_id, inference_class))
  train, val, test = dataloader(tokenizer.encode)
  print(f"Evaluating {len(test)} examples with batch_size {batch_size}")
  loss, tokens = await run_iter(node, shard, False, test, batch_size)
  print(f"total | {loss=}, {tokens=}")
  print("Waiting for outstanding tasks")
  await hold_outstanding(node)

async def train_model_cli(node: Node, model_name, dataloader, batch_size, iters, save_interval=0, checkpoint_dir=None):
  inference_class = node.inference_engine.__class__.__name__
  shard = build_base_shard(model_name, inference_class)
  if not shard:
    print(f"Error: Unsupported model '{model_name}' for inference engine {inference_class}")
    return
  tokenizer = await resolve_tokenizer(get_repo(shard.model_id, inference_class))
  train, val, test = dataloader(tokenizer.encode)
  print(f"Training on {len(train)} examples with batch_size {batch_size} for {iters} epochs")
  for i in tqdm(range(3)):
    await asyncio.sleep(1)
  for epoch in range(iters):
    loss, tokens = await run_iter(node, shard, True, train, batch_size)
    print(f"epoch {epoch + 1}/{iters}\t| loss: {loss}, tokens: {tokens}")
    if save_interval > 0 and epoch > 0 and (epoch % save_interval) == 0 and checkpoint_dir is not None:
      await node.coordinate_save(shard, epoch, checkpoint_dir)
      await hold_outstanding(node)
  await hold_outstanding(node)

async def check_exo_home():
  home, has_read, has_write = await ensure_exo_home(), await has_exo_home_read_access(), await has_exo_home_write_access()
  if DEBUG >= 1: print(f"exo home directory: {home}")
  print(f"{has_read=}, {has_write=}")
  if not has_read or not has_write:
    print(f"""
          WARNING: Limited permissions for exo home directory: {home}.
          This may prevent model downloads from working correctly.
          {"❌ No read access" if not has_read else ""}
          {"❌ No write access" if not has_write else ""}
          """)

async def main():
  loop = asyncio.get_running_loop()

  try: await check_exo_home()
  except Exception as e: print(f"Error checking exo home directory: {e}")

  if not args.models_seed_dir is None:
    try:
      models_seed_dir = clean_path(args.models_seed_dir)
      await seed_models(models_seed_dir)
    except Exception as e:
      print(f"Error seeding models: {e}")

  def restore_cursor():
    if platform.system() != "Windows":
        os.system("tput cnorm")  # Show cursor

  # Restore the cursor when the program exits
  atexit.register(restore_cursor)

  # Use a more direct approach to handle signals
  shutdown_event = asyncio.Event()
  
  def handle_exit(sig):
    # Don't print the stack trace on Ctrl-C
    if sig == signal.SIGINT:
      print()  # New line after ^C
    shutdown_event.set()
    # Create task for graceful shutdown
    asyncio.create_task(shutdown(sig, loop, node.server))

  if platform.system() != "Windows":
    for s in [signal.SIGINT, signal.SIGTERM]:
      loop.add_signal_handler(s, lambda: handle_exit(s))

  await node.start(wait_for_peers=args.wait_for_peers)

  if args.command == "run" or args.run_model:
    model_name = args.model_name or args.run_model
    if not model_name:
      print("Error: Model name is required when using 'run' command or --run-model")
      return
    await run_model_cli(node, model_name, args.prompt)
  elif args.command == "eval" or args.command == 'train':
    model_name = args.model_name
    dataloader = lambda tok: load_dataset(args.data, preprocess=lambda item: tok(item)
                                                   , loadline=lambda line: json.loads(line).get("text",""))
    if args.command == 'eval':
      if not model_name:
        print("Error: Much like a human, I can't evaluate anything without a model")
        return
      await eval_model_cli(node, model_name, dataloader, args.batch_size)
    else:
      if not model_name:
        print("Error: This train ain't leaving the station without a model")
        return
      await train_model_cli(node, model_name, dataloader, args.batch_size, args.iters, save_interval=args.save_every, checkpoint_dir=args.save_checkpoint_dir)

  else:
    asyncio.create_task(api.run(port=args.chatgpt_api_port))  # Start the API server as a non-blocking task
    await asyncio.Event().wait()

  if args.wait_for_peers > 0:
    print("Cooldown to allow peers to exit gracefully")
    for i in tqdm(range(50)):
      await asyncio.sleep(.1)

def run():
    """Run the main function, handling event loop setup."""
    # Check if we're already in an async context
    try:
        loop = asyncio.get_running_loop()
        # We're already in an async context, just return the coroutine
        return main()
    except RuntimeError:
        # No running loop, create one
        loop = None
        try:
            loop = configure_uvloop()
            loop.run_until_complete(main())
        except KeyboardInterrupt:
            pass  # Handled by signal handler
        except Exception as e:
            print(f"Error: {e}")
        finally:
            if loop and not loop.is_closed(): 
                try:
                    loop.close()
                except:
                    pass

if __name__ == "__main__":
  # When run directly, use synchronous version
  loop = None
  try:
      loop = configure_uvloop()
      loop.run_until_complete(main())
  except KeyboardInterrupt:
      pass  # Handled by signal handler
  except Exception as e:
      print(f"Error: {e}")
  finally:
      if loop and not loop.is_closed():
          try:
              loop.close()
          except:
              pass



import unittest
from typing import List
from net.topology.partitioning_strategy import Partition, map_partitions_to_shards
from net.inference.shard import Shard


class TestRingMemoryWeightedPartitioningStrategy(unittest.TestCase):
  def test_map_partitions_to_shards(self):
    partitions = [
      Partition("node1", 0.0, 0.42857),
      Partition("node2", 0.42857, 0.71428),
      Partition("node3", 0.71428, 0.99999),
    ]
    shards = map_partitions_to_shards(partitions, 32, "model")
    self.assertEqual(
      shards,
      [
        Shard("model", 0, 12, 32),
        Shard("model", 13, 21, 32),
        Shard("model", 22, 31, 32),
      ],
    )

    partitions = [
      Partition("node1", 0.0, 0.1),
      Partition("node2", 0.1, 0.2),
      Partition("node3", 0.2, 1.0),
    ]
    shards = map_partitions_to_shards(partitions, 32, "model")
    self.assertEqual(
      shards,
      [
        Shard("model", 0, 2, 32),
        Shard("model", 3, 5, 32),
        Shard("model", 6, 31, 32),
      ],
    )

    partitions = [
      Partition("node1", 0.0, 1.0),
    ]
    shards = map_partitions_to_shards(partitions, 32, "model")
    self.assertEqual(
      shards,
      [
        Shard("model", 0, 31, 32),
      ],
    )

    partitions = []
    shards = map_partitions_to_shards(partitions, 32, "model")
    self.assertEqual(shards, [])

  def test_broken_map_partitions_to_shards(self):
    # this was an old broken implementation that sometimes had rounding errors!
    def _broken_map_partitions_to_shards(partitions: List[Partition], num_layers, model_id: str):
      shards = []
      for i, partition in enumerate(partitions):
        start_layer = int(partition.start*num_layers)
        end_layer = int(partition.end*num_layers) - 1
        shards.append(Shard(model_id, start_layer, end_layer, num_layers))
      return shards

    partitions = [
      Partition("node1", 0.0, 0.42857),
      Partition("node2", 0.42857, 0.71428),
      Partition("node3", 0.71428, 0.99999),
    ]
    shards = _broken_map_partitions_to_shards(partitions, 32, "model")
    self.assertEqual(
      shards,
      [
        Shard("model", 0, 12, 32),
        Shard("model", 13, 21, 32),
        Shard("model", 22, 30, 32),
      ],
    )


if __name__ == "__main__":
  unittest.main()



import unittest

from net.networking.manual.network_topology_config import NetworkTopology

root_path = "./exo/networking/manual/test_data/"


class TestNetworkTopologyConfig(unittest.TestCase):
  def test_from_path_invalid_path(self):
    with self.assertRaises(FileNotFoundError) as e:
      NetworkTopology.from_path("invalid_path")
    self.assertEqual(str(e.exception), "Config file not found at invalid_path")

  def test_from_path_invalid_json(self):
    with self.assertRaises(ValueError) as e:
      NetworkTopology.from_path(root_path + "invalid_json.json")
    self.assertIn("Error validating network topology config from", str(e.exception))
    self.assertIn("1 validation error for NetworkTopology\n  Invalid JSON: EOF while parsing a value at line 1 column 0", str(e.exception))

  def test_from_path_invalid_config(self):
    with self.assertRaises(ValueError) as e:
      NetworkTopology.from_path(root_path + "invalid_config.json")
    self.assertIn("Error validating network topology config from", str(e.exception))
    self.assertIn("port\n  Field required", str(e.exception))

  def test_from_path_valid(self):
    config = NetworkTopology.from_path(root_path + "test_config.json")

    self.assertEqual(config.peers["node1"].port, 50051)
    self.assertEqual(config.peers["node1"].device_capabilities.model, "Unknown Model")
    self.assertEqual(config.peers["node1"].address, "localhost")
    self.assertEqual(config.peers["node1"].device_capabilities.chip, "Unknown Chip")
    self.assertEqual(config.peers["node1"].device_capabilities.memory, 0)
    self.assertEqual(config.peers["node1"].device_capabilities.flops.fp32, 0)
    self.assertEqual(config.peers["node1"].device_capabilities.flops.fp16, 0)
    self.assertEqual(config.peers["node1"].device_capabilities.flops.int8, 0)

    self.assertEqual(config.peers["node2"].port, 50052)
    self.assertEqual(config.peers["node2"].device_capabilities.model, "Unknown Model")
    self.assertEqual(config.peers["node2"].address, "localhost")
    self.assertEqual(config.peers["node2"].device_capabilities.chip, "Unknown Chip")
    self.assertEqual(config.peers["node2"].device_capabilities.memory, 0)
    self.assertEqual(config.peers["node2"].device_capabilities.flops.fp32, 0)
    self.assertEqual(config.peers["node2"].device_capabilities.flops.fp16, 0)
    self.assertEqual(config.peers["node2"].device_capabilities.flops.int8, 0)


if __name__ == "__main__":
  unittest.main()



from PIL import Image, ImageDraw, ImageFont, ImageFilter
import os
import numpy as np
import cv2
import sys

def draw_rounded_rectangle(draw, coords, radius, fill):
  left, top, right, bottom = coords
  diameter = radius * 2
  draw.rectangle([left + radius, top, right - radius, bottom], fill=fill)
  draw.rectangle([left, top + radius, right, bottom - radius], fill=fill)
  draw.pieslice([left, top, left + diameter, top + diameter], 180, 270, fill=fill)
  draw.pieslice([right - diameter, top, right, top + diameter], 270, 360, fill=fill)
  draw.pieslice([left, bottom - diameter, left + diameter, bottom], 90, 180, fill=fill)
  draw.pieslice([right - diameter, bottom - diameter, right, bottom], 0, 90, fill=fill)

def draw_centered_text_rounded(draw, text, font, rect_coords, radius=10, text_color="yellow", bg_color=(43,33,44)):
  bbox = font.getbbox(text)
  text_width = bbox[2] - bbox[0]
  text_height = bbox[3] - bbox[1]
  rect_left, rect_top, rect_right, rect_bottom = rect_coords
  rect_width = rect_right - rect_left
  rect_height = rect_bottom - rect_top
  text_x = rect_left + (rect_width - text_width) // 2
  text_y = rect_top + (rect_height - text_height) // 2
  draw_rounded_rectangle(draw, rect_coords, radius, bg_color)
  draw.text((text_x, text_y), text, fill=text_color, font=font)

def draw_left_aligned_text_rounded(draw, text, font, rect_coords, padding_left=20, radius=10, text_color="yellow", bg_color=(43,33,44)):
  bbox = font.getbbox(text)
  text_height = bbox[3] - bbox[1]
  rect_left, rect_top, rect_right, rect_bottom = rect_coords
  rect_height = rect_bottom - rect_top
  text_y = rect_top + (rect_height - text_height) // 2
  text_x = rect_left + padding_left
  draw_rounded_rectangle(draw, rect_coords, radius, bg_color)
  draw.text((text_x, text_y), text, fill=text_color, font=font)

def draw_right_text_dynamic_width_rounded(draw, text, font, base_coords, padding=20, radius=10, text_color="yellow", bg_color=(43,33,44)):
  bbox = font.getbbox(text)
  text_width = bbox[2] - bbox[0]
  text_height = bbox[3] - bbox[1]
  _, rect_top, rect_right, rect_bottom = base_coords
  rect_height = rect_bottom - rect_top
  new_rect_left = rect_right - (text_width + (padding * 2))
  text_y = rect_top + (rect_height - text_height) // 2
  text_x = new_rect_left + padding
  draw_rounded_rectangle(draw, (new_rect_left, rect_top, rect_right, rect_bottom), radius, bg_color)
  draw.text((text_x, text_y), text, fill=text_color, font=font)
  return new_rect_left

def draw_progress_bar(draw, progress, coords, color="yellow", bg_color=(70, 70, 70)):
  left, top, right, bottom = coords
  total_width = right - left
  draw.rectangle(coords, fill=bg_color)
  progress_width = int(total_width * progress)
  if progress_width > 0:
    draw.rectangle((left, top, left + progress_width, bottom), fill=color)

def crop_image(image, top_crop=70):
  width, height = image.size
  return image.crop((0, top_crop, width, height))

def create_animation_mp4(
  replacement_image_path,
  output_path,
  device_name,
  prompt_text,
  fps=30,
  target_size=(512, 512),
  target_position=(139, 755),
  progress_coords=(139, 1285, 655, 1295),
  device_coords=(1240, 370, 1640, 416),
  prompt_coords=(332, 1702, 2662, 1745)
):
  frames = []
  try:
    font = ImageFont.truetype("/System/Library/Fonts/SFNSMono.ttf", 20)
    promptfont = ImageFont.truetype("/System/Library/Fonts/SFNSMono.ttf", 24)
  except:
    font = ImageFont.load_default()
    promptfont = ImageFont.load_default()

  # Get the base directory for images when running as a bundled app
  if hasattr(sys, '_MEIPASS'):
    base_dir = os.path.join(sys._MEIPASS, "exo", "apputil", "baseimages")
  else:
    base_dir = os.path.join(os.path.dirname(__file__), "baseimages")

  # Process first frame
  base_img = Image.open(os.path.join(base_dir, "image1.png"))
  draw = ImageDraw.Draw(base_img)
  draw_centered_text_rounded(draw, device_name, font, device_coords)
  frames.extend([crop_image(base_img)] * 30)  # 1 second at 30fps

  # Process second frame with typing animation
  base_img2 = Image.open(os.path.join(base_dir, "image2.png"))
  for i in range(len(prompt_text) + 1):
    current_frame = base_img2.copy()
    draw = ImageDraw.Draw(current_frame)
    draw_centered_text_rounded(draw, device_name, font, device_coords)
    if i > 0:  # Only draw if we have at least one character
      draw_left_aligned_text_rounded(draw, prompt_text[:i], promptfont, prompt_coords)
    frames.extend([crop_image(current_frame)] * 2)  # 2 frames per character for smooth typing
  
  # Hold the complete prompt for a moment
  frames.extend([frames[-1]] * 30)  # Hold for 1 second

  # Create blur sequence
  replacement_img = Image.open(replacement_image_path)
  base_img = Image.open(os.path.join(base_dir, "image3.png"))
  blur_steps = [int(80 * (1 - i/8)) for i in range(9)]

  for i, blur_amount in enumerate(blur_steps):
    new_frame = base_img.copy()
    draw = ImageDraw.Draw(new_frame)

    replacement_copy = replacement_img.copy()
    replacement_copy.thumbnail(target_size, Image.Resampling.LANCZOS)
    if blur_amount > 0:
      replacement_copy = replacement_copy.filter(ImageFilter.GaussianBlur(radius=blur_amount))

    mask = replacement_copy.split()[-1] if replacement_copy.mode in ('RGBA', 'LA') else None
    new_frame.paste(replacement_copy, target_position, mask)

    draw_progress_bar(draw, (i + 1) / 9, progress_coords)
    draw_centered_text_rounded(draw, device_name, font, device_coords)
    draw_right_text_dynamic_width_rounded(draw, prompt_text, promptfont, (None, 590, 2850, 685), padding=30)

    frames.extend([crop_image(new_frame)] * 15)  # 0.5 seconds at 30fps

  # Create and add final frame (image4)
  final_base = Image.open(os.path.join(base_dir, "image4.png"))
  draw = ImageDraw.Draw(final_base)

  draw_centered_text_rounded(draw, device_name, font, device_coords)
  draw_right_text_dynamic_width_rounded(draw, prompt_text, promptfont, (None, 590, 2850, 685), padding=30)

  replacement_copy = replacement_img.copy()
  replacement_copy.thumbnail(target_size, Image.Resampling.LANCZOS)
  mask = replacement_copy.split()[-1] if replacement_copy.mode in ('RGBA', 'LA') else None
  final_base.paste(replacement_copy, target_position, mask)

  frames.extend([crop_image(final_base)] * 30)  # 1 second at 30fps

  # Convert frames to video using H.264 codec
  if frames:
    first_frame = np.array(frames[0])
    height, width = first_frame.shape[:2]
    fourcc = cv2.VideoWriter_fourcc(*'avc1')
    out = cv2.VideoWriter(
      output_path,
      fourcc,
      fps,
      (width, height),
      isColor=True
    )

    if not out.isOpened():
      print("Error: VideoWriter failed to open")
      return

    for frame in frames:
      frame_array = cv2.cvtColor(np.array(frame), cv2.COLOR_RGB2BGR)
      out.write(frame_array)
    
    out.release()
    print(f"Video saved successfully to {output_path}")



from dataclasses import dataclass, field

import mlx.core as mx
import mlx.nn as nn

from mlx_lm.models.base import create_attention_mask
from mlx_lm.models.phi3 import TransformerBlock, ModelArgs

from ...shard import Shard
from .base import IdentityBlock

@dataclass
class ModelArgs(ModelArgs):
  shard: Shard = field(default_factory=lambda: Shard("", 0, 0, 0))

  def __post_init__(self):
    super().__post_init__()

    if isinstance(self.shard, Shard):
      return
    if not isinstance(self.shard, dict):
      raise TypeError(f"Expected shard to be a Shard instance or a dict, got {type(self.shard)} instead")

    self.shard = Shard(**self.shard)

class Phi3Model(nn.Module):
  def __init__(self, args: ModelArgs):
    super().__init__()
    self.args = args
    self.vocab_size = args.vocab_size
    self.num_hidden_layers = args.num_hidden_layers
    assert self.vocab_size > 0
    
    if self.args.shard.is_first_layer():
      self.embed_tokens = nn.Embedding(args.vocab_size, args.hidden_size)
    
    self.layers = []
    for i in range(self.num_hidden_layers):
      if self.args.shard.start_layer <= i <= self.args.shard.end_layer:
        self.layers.append(TransformerBlock(args=args))
      else:
        self.layers.append(IdentityBlock())
        
    if self.args.shard.is_last_layer():
      self.norm = nn.RMSNorm(args.hidden_size, eps=args.rms_norm_eps)

  def __call__(
    self,
    inputs: mx.array,
    cache=None,
  ):
    if self.args.shard.is_first_layer():
      h = self.embed_tokens(inputs)
    else:
      h = inputs

    mask = None
    if h.shape[1] > 1:
      mask = create_attention_mask(h, cache)

    if cache is None:
      cache = [None] * len(self.layers)

    for layer, c in zip(self.layers, cache):
      h = layer(h, mask, c)

    if self.args.shard.is_last_layer():
      h = self.norm(h)
    return h

class Model(nn.Module):
  def __init__(self, args: ModelArgs):
    super().__init__()
    self.args = args
    self.model_type = args.model_type
    self.model = Phi3Model(args)
    if self.args.shard.is_last_layer():
      self.lm_head = nn.Linear(args.hidden_size, args.vocab_size, bias=False)

  def __call__(
    self,
    inputs: mx.array,
    cache=None,
  ):
    out = self.model(inputs, cache)
    if self.args.shard.is_last_layer():
      out = self.lm_head(out)
    return out

  def sanitize(self, weights):
    shard_state_dict = {}

    for key, value in weights.items():
      if "self_attn.rope.inv_freq" in key:
        continue
      if key.startswith('model.layers.'):
        layer_num = int(key.split('.')[2])
        if self.args.shard.start_layer <= layer_num <= self.args.shard.end_layer:
          shard_state_dict[key] = value
      elif self.args.shard.is_first_layer() and key.startswith('model.embed_tokens'):
        shard_state_dict[key] = value
      elif self.args.shard.is_last_layer() and (key.startswith('lm_head') or key.startswith('model.norm')):
        shard_state_dict[key] = value

    return shard_state_dict

  @property
  def layers(self):
    return self.model.layers

  @property
  def head_dim(self):
    return self.args.hidden_size // self.args.num_attention_heads

  @property
  def n_kv_heads(self):
    return self.args.num_key_value_heads



from .device_capabilities import DeviceCapabilities
from typing import Dict, Set, Optional
from dataclasses import dataclass

@dataclass
class PeerConnection:
  from_id: str
  to_id: str
  description: Optional[str] = None

  def __hash__(self):
    # Use both from_id and to_id for uniqueness in sets
    return hash((self.from_id, self.to_id))

  def __eq__(self, other):
    if not isinstance(other, PeerConnection):
      return False
    # Compare both from_id and to_id for equality
    return self.from_id == other.from_id and self.to_id == other.to_id

class Topology:
  def __init__(self):
    self.nodes: Dict[str, DeviceCapabilities] = {}
    self.peer_graph: Dict[str, Set[PeerConnection]] = {}
    self.active_node_id: Optional[str] = None

  def update_node(self, node_id: str, device_capabilities: DeviceCapabilities):
    self.nodes[node_id] = device_capabilities

  def get_node(self, node_id: str) -> DeviceCapabilities:
    return self.nodes.get(node_id)

  def all_nodes(self):
    return self.nodes.items()

  def add_edge(self, from_id: str, to_id: str, description: Optional[str] = None):
    if from_id not in self.peer_graph:
      self.peer_graph[from_id] = set()
    conn = PeerConnection(from_id, to_id, description)
    self.peer_graph[from_id].add(conn)

  def merge(self, peer_node_id: str, other: "Topology"):
    for node_id, capabilities in other.nodes.items():
      if node_id != peer_node_id: continue
      self.update_node(node_id, capabilities)
    for node_id, connections in other.peer_graph.items():
      for conn in connections:
        if conn.from_id != peer_node_id: continue
        self.add_edge(conn.from_id, conn.to_id, conn.description)

  def __str__(self):
    nodes_str = ", ".join(f"{node_id}: {cap}" for node_id, cap in self.nodes.items())
    edges_str = ", ".join(f"{node}: {[f'{c.to_id}({c.description})' for c in conns]}"
                         for node, conns in self.peer_graph.items())
    return f"Topology(Nodes: {{{nodes_str}}}, Edges: {{{edges_str}}})"

  def to_json(self):
    return {
      "nodes": {
        node_id: capabilities.to_dict()
        for node_id, capabilities in self.nodes.items()
      },
      "peer_graph": {
        node_id: [
          {
            "from_id": conn.from_id,
            "to_id": conn.to_id,
            "description": conn.description
          }
          for conn in connections
        ]
        for node_id, connections in self.peer_graph.items()
      },
      "active_node_id": self.active_node_id
    }
